"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const path_1 = tslib_1.__importDefault(require("path"));
const constants_1 = require("../constants");
const alreadyTriedSrcPath = Symbol('alreadyTriedSrcPath');
/**
 * 获取对象字段值
 * @param content 对象
 * @param field 字段名称，可以是一个数组
 * @returns 字段值
 */
function getObjectFieldValue(content, field) {
    if (!content)
        return undefined;
    if (Array.isArray(field)) {
        let current;
        for (let j = 0; j < field.length; j++) {
            current = content === null || content === void 0 ? void 0 : content[field[j]];
            if (current)
                return current;
        }
        return current;
    }
    else {
        return content === null || content === void 0 ? void 0 : content[field];
    }
}
function forEachBail(array, iterator, callback) {
    if (array.length === 0)
        return callback();
    let i = 0;
    const next = () => {
        let loop = undefined;
        iterator(array[i++], (err, result) => {
            if (err || result !== undefined || i >= array.length) {
                return callback(err, result);
            }
            if (loop === false)
                while (next())
                    ;
            loop = true;
        });
        if (!loop)
            loop = false;
        return loop;
    };
    while (next())
        ;
}
/**
 * 排序之后的源码路径
 * @param srcPaths - 源码路径
 * @param request - 请求的文件
 * @returns 返回排序后的源码目录列表和相对路径
 */
function sortedSrcPaths(srcPaths, requestPath) {
    const srcs = [];
    let relativePath;
    for (const src of srcPaths || []) {
        if (requestPath.startsWith(src)) {
            // 尝试获取当前文件相对于 srcPath 的路径
            const relative = path_1.default.relative(src, requestPath);
            if (!relativePath || relativePath.length > relative.length) {
                relativePath = relative;
                srcs.unshift(src);
            }
            else {
                srcs.push(src);
            }
        }
        else {
            srcs.push(src);
        }
    }
    return { srcs, relativePath };
}
/**
 * webpack enhance-resolver 插件
 * 1. 将绝对路径的引用限制在 srcPath 之内
 * 2. 支持多端产物目录检索
 */
class ResolverPlugin {
    constructor(options) {
        this.name = 'MorResolverPlugin';
        this.options = { ...options };
    }
    apply(resolver) {
        this.restrictToAndAllowMultipleSrcPaths(resolver);
        this.searchInMultiDist(resolver);
    }
    /**
     * 限制绝对路径文件查找 以及 允许多个 srcPaths 查询
     */
    restrictToAndAllowMultipleSrcPaths(resolver) {
        const { srcPaths } = this.options;
        const parsedResolveTarget = resolver.ensureHook('parsed-resolve');
        const directoryTarget = resolver.ensureHook('directory');
        const rawFileTarget = resolver.ensureHook('raw-file');
        /**
         * 绝对路径查询支持
         * 1. 优先查询当前文件所在 srcPath 的绝对路径
         * 2. 然后再查询其他 srcPaths 中的文件
         */
        resolver
            .getHook('resolve')
            .tapAsync(this.name, (request, resolveContext, callback) => {
            var _a;
            const parsed = resolver.parse(request.request);
            // 检查条件
            // 请求的文件(request) 为绝对路径, 且不在任意一个 src 中
            // 且存在 issuer
            if (request.request &&
                path_1.default.isAbsolute(request.request) &&
                request.path &&
                ((_a = request['context']) === null || _a === void 0 ? void 0 : _a.issuer)) {
                const { srcs, relativePath } = sortedSrcPaths(srcPaths, request.request);
                // 如果存在相对路径, 则代表当前请求的绝对路径位于某个 srcPath 中
                if (relativePath)
                    return callback();
                forEachBail(srcs, (srcPath, clk) => {
                    const obj = { ...request, ...parsed };
                    if (obj.request.startsWith(srcPath)) {
                        obj.request = path_1.default.resolve(srcPath, obj.request);
                    }
                    else {
                        obj.request = path_1.default.join(srcPath, obj.request);
                    }
                    resolver.doResolve(parsedResolveTarget, obj, 'try resolve absolute path in ' + srcPath, resolveContext, clk);
                }, callback);
            }
            else {
                callback();
            }
        });
        /**
         * 支持多路径检索
         * 优先检索当前目录的路径, 然后扩展到其他源码目录中
         */
        resolver
            .getHook('described-relative')
            .tapAsync(this.name, (request, resolveContext, callback) => {
            if (request.request || !request.path)
                return callback();
            // 不处理 node_modules 中的路径
            if (constants_1.NODE_MODULE_REGEXP.test(request.path))
                return callback();
            const { srcs, relativePath } = sortedSrcPaths(srcPaths, request.path);
            // 如果未获取到相对路径，则代表当前文件不在任何一个 srcPath 中
            // 不做任何处理
            if (relativePath == null)
                return callback();
            forEachBail(srcs, (srcPath, clk) => {
                if (request[alreadyTriedSrcPath] === srcPath)
                    return clk();
                // 使用新的 srcPath 组装备选文件路径
                const alternativePath = path_1.default.join(srcPath, relativePath);
                resolver.doResolve(request.directory ? directoryTarget : rawFileTarget, {
                    ...request,
                    [alreadyTriedSrcPath]: srcPath,
                    path: alternativePath,
                    relativePath: request.descriptionFileRoot
                        ? path_1.default.relative(request.descriptionFileRoot, alternativePath)
                        : undefined,
                    request: undefined
                }, 'try load file in alternative srcPath ' + srcPath, resolveContext, clk);
            }, callback);
        });
    }
    /**
     * 多端编译的 npm 查找支持, 适配 微信的 miniprogram_dist 逻辑
     */
    searchInMultiDist(resolver) {
        var _a;
        const { defaultNpmMiniProgramDist } = this.options;
        // 小程序多端 mainFields 对应的组件查询支持
        const target = resolver.ensureHook('resolve-in-existing-directory');
        const mainFields = ((_a = resolver === null || resolver === void 0 ? void 0 : resolver.options) === null || _a === void 0 ? void 0 : _a.mainFields) || [];
        // 合并 mainFields
        // 需要考虑 mainFields 可能是一个文件的情况
        // 另外如果一个 mainField 没有找到文件，需要能够falback到优先级更低的 mainField
        const allMainFields = mainFields.reduce(function (result, item) {
            return result.concat(item.name);
        }, []);
        resolver
            .getHook('resolve-in-package')
            .tapAsync(this.name, (request, resolveContext, callback) => {
            // 不处理 node_modules 以外的逻辑
            if (!request.descriptionFilePath)
                return callback();
            if (!request.descriptionFileData)
                return callback();
            // package.json 文件可能继承自 父 package
            // 代表当前 package 无 package.json 文件, 不处理这种情况
            if (request.relativePath !== '.')
                return callback();
            // 不处理 request 为空 或 以 .. 起始的情况
            if (!request.request || request.request.startsWith('..')) {
                return callback();
            }
            let multiDist = getObjectFieldValue(request.descriptionFileData, allMainFields) ||
                defaultNpmMiniProgramDist;
            // 如果无值 或者 值不是字符串, 则忽略
            if (!multiDist || typeof multiDist !== 'string')
                return callback();
            // 如果配置的不是文件夹，而是一个文件
            if (path_1.default.extname(multiDist) !== '') {
                // 若默认产物目录存在，则优先使用
                if (defaultNpmMiniProgramDist) {
                    multiDist = defaultNpmMiniProgramDist;
                }
                else {
                    // 否则, 不处理
                    return callback();
                }
            }
            // 如果路径已经以 multiDist 作为起始, 则不处理
            if (request.request.startsWith(multiDist) ||
                request.request.startsWith('./' + multiDist)) {
                return callback();
            }
            resolver.doResolve(target, {
                ...request,
                request: path_1.default.join(multiDist, request.request)
            }, 'try load file in ' + multiDist, resolveContext, callback);
        });
    }
}
exports.default = ResolverPlugin;
//# sourceMappingURL=resolverPlugin.js.map